@@ -95,12 +95,14 @@ enum {
 	PROP_RIGHT_MARGIN_POSITION,
 	PROP_SMART_HOME_END,
 	PROP_HIGHLIGHT_CURRENT_LINE,
-	PROP_INDENT_ON_TAB
+	PROP_INDENT_ON_TAB,
+	PROP_DRAW_SPACES
 };
 
 struct _GtkSourceViewPrivate
 {
 	guint		 tab_width;
+	gboolean	 tabs_set;
 	gint		 indent_width;
 	gboolean 	 show_line_numbers;
 	gboolean	 show_line_marks;
@@ -116,14 +118,19 @@ struct _GtkSourceViewPrivate
 
 	gboolean	 style_scheme_applied;
 	GtkSourceStyleScheme *style_scheme;
-	GdkGC		*current_line_gc;
 	GdkColor        *right_margin_line_color;
 	GdkColor        *right_margin_overlay_color;
 
+	GtkSourceDrawSpacesFlags draw_spaces;
+	GdkColor        *spaces_color;
+
 	GHashTable 	*mark_categories;
 
 	GtkSourceBuffer *source_buffer;
 	gint		 old_lines;
+
+	GdkColor         current_line_color;
+	guint            current_line_color_set : 1;
 };
 
 
@@ -141,10 +148,12 @@ static const GtkTargetEntry drop_types[] = {
 
 static guint signals[LAST_SIGNAL] = { 0 };
 
-typedef struct 
+typedef struct
 {
 	gint priority;
 	GdkPixbuf *pixbuf;
+	GdkColor background;
+	guint background_set;
 } MarkCategory;
 
 /* Prototypes. */
@@ -174,6 +183,7 @@ static void 	gtk_source_view_get_lines 		(GtkTextView       *text_view,
 				       			 gint               first_y,
 				       			 gint               last_y,
 				       			 GArray            *buffer_coords,
+				       			 GArray            *line_heights,
 				       			 GArray            *numbers,
 				       			 gint              *countp);
 static gint     gtk_source_view_expose 			(GtkWidget         *widget,
@@ -207,7 +217,6 @@ static void	gtk_source_view_get_property		(GObject           *object,
 static void     gtk_source_view_style_set               (GtkWidget         *widget,
 							 GtkStyle          *previous_style);
 static void	gtk_source_view_realize			(GtkWidget         *widget);
-static void	gtk_source_view_unrealize		(GtkWidget         *widget);
 static void	gtk_source_view_update_style_scheme	(GtkSourceView     *view);
 
 static MarkCategory *
@@ -238,14 +247,13 @@ gtk_source_view_class_init (GtkSourceViewClass *klass)
 	widget_class->expose_event = gtk_source_view_expose;
 	widget_class->style_set = gtk_source_view_style_set;
 	widget_class->realize = gtk_source_view_realize;
-	widget_class->unrealize = gtk_source_view_unrealize;
 
 	textview_class->populate_popup = gtk_source_view_populate_popup;
 	textview_class->move_cursor = gtk_source_view_move_cursor;
 
 	klass->undo = gtk_source_view_undo;
 	klass->redo = gtk_source_view_redo;
-	
+
 	/**
 	 * GtkSourceView:show-line-numbers:
 	 *
@@ -379,6 +387,22 @@ gtk_source_view_class_init (GtkSourceViewClass *klass)
 							       TRUE,
 							       G_PARAM_READWRITE));
 
+	/**
+	 * GtkSourceView:draw-spaces:
+	 *
+	 * Set if and how the spaces should be visualized.
+	 *
+	 * Since: 2.4
+	 */
+	g_object_class_install_property (object_class,
+					 PROP_DRAW_SPACES,
+					 g_param_spec_flags ("draw-spaces",
+							    _("Draw Spaces"),
+							    _("Set if and how the spaces should be visualized"),
+							    GTK_TYPE_SOURCE_DRAW_SPACES_FLAGS,
+							    0,
+							    G_PARAM_READWRITE));
+
 	signals [UNDO] =
 		g_signal_new ("undo",
 			      G_TYPE_FROM_CLASS (klass),
@@ -572,6 +596,11 @@ gtk_source_view_set_property (GObject      *object,
 							   g_value_get_boolean (value));
 			break;
 
+		case PROP_DRAW_SPACES:
+			gtk_source_view_set_draw_spaces (view,
+							 g_value_get_flags (value));
+			break;
+
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 			break;
@@ -633,8 +662,8 @@ gtk_source_view_get_property (GObject    *object,
 			break;
 
 		case PROP_SMART_HOME_END:
-			g_value_set_flags (value,
-					   gtk_source_view_get_smart_home_end (view));
+			g_value_set_enum (value,
+					  gtk_source_view_get_smart_home_end (view));
 			break;
 
 		case PROP_HIGHLIGHT_CURRENT_LINE:
@@ -647,6 +676,11 @@ gtk_source_view_get_property (GObject    *object,
 					     gtk_source_view_get_indent_on_tab (view));
 			break;
 
+		case PROP_DRAW_SPACES:
+			g_value_set_flags (value,
+					   gtk_source_view_get_draw_spaces (view));
+			break;
+
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 			break;
@@ -671,6 +705,7 @@ gtk_source_view_init (GtkSourceView *view)
 						  GtkSourceViewPrivate);
 
 	view->priv->tab_width = DEFAULT_TAB_WIDTH;
+	view->priv->tabs_set = FALSE;
 	view->priv->indent_width = -1;
 	view->priv->indent_on_tab = TRUE;
 	view->priv->smart_home_end = GTK_SOURCE_SMART_HOME_END_DISABLED;
@@ -682,6 +717,7 @@ gtk_source_view_init (GtkSourceView *view)
 
 	view->priv->right_margin_line_color = NULL;
 	view->priv->right_margin_overlay_color = NULL;
+	view->priv->spaces_color = NULL;
 
 	view->priv->mark_categories = g_hash_table_new_full (g_str_hash, g_str_equal,
 							     (GDestroyNotify) g_free,
@@ -733,12 +769,15 @@ gtk_source_view_finalize (GObject *object)
 
 	if (view->priv->style_scheme)
 		g_object_unref (view->priv->style_scheme);
-	
+
 	if (view->priv->right_margin_line_color != NULL)
 		gdk_color_free (view->priv->right_margin_line_color);
 
 	if (view->priv->right_margin_overlay_color != NULL)
 		gdk_color_free (view->priv->right_margin_overlay_color);
+		
+	if (view->priv->spaces_color != NULL)
+		gdk_color_free (view->priv->spaces_color);
 
 	if (view->priv->mark_categories)
 		g_hash_table_destroy (view->priv->mark_categories);
@@ -819,16 +858,9 @@ source_mark_updated_cb (GtkSourceBuffer *buffer,
 			GtkSourceMark	*mark,
 			GtkTextView     *text_view)
 {
-	GdkWindow *margin;
-
-	margin = gtk_text_view_get_window (text_view, GTK_TEXT_WINDOW_LEFT);
-	if (margin != NULL)
-	{
-		GdkRegion *region;
-
-		region = gdk_drawable_get_visible_region (GDK_DRAWABLE (margin));
-		gdk_window_invalidate_region (margin, region, TRUE);
-	}
+	/* TODO do something more intelligent here, namely
+	 * invalidate only the area under the mark if possible */
+	gtk_widget_queue_draw (GTK_WIDGET (text_view));
 }
 
 static void
@@ -882,6 +914,9 @@ set_source_buffer (GtkSourceView *view,
 		view->priv->source_buffer = NULL;
 	}
 
+	/* if buffer isn't NULL then we aren't being destroyed, so call
+	 * gtk_source_view_update_style_scheme(), that will check whether it's
+	 * a GtkSourceBuffer or not */
 	if (buffer)
 		gtk_source_view_update_style_scheme (view);
 }
@@ -896,6 +931,7 @@ gtk_source_view_undo (GtkSourceView *view)
 	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (view));
 
 	if (gtk_text_view_get_editable (GTK_TEXT_VIEW (view)) &&
+	    GTK_IS_SOURCE_BUFFER (buffer) &&
 	    gtk_source_buffer_can_undo (GTK_SOURCE_BUFFER (buffer)))
 	{
 		gtk_source_buffer_undo (GTK_SOURCE_BUFFER (buffer));
@@ -914,6 +950,7 @@ gtk_source_view_redo (GtkSourceView *view)
 	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (view));
 
 	if (gtk_text_view_get_editable (GTK_TEXT_VIEW (view)) &&
+	    GTK_IS_SOURCE_BUFFER (buffer) &&
 	    gtk_source_buffer_can_redo (GTK_SOURCE_BUFFER (buffer)))
 	{
 		gtk_source_buffer_redo (GTK_SOURCE_BUFFER (buffer));
@@ -930,7 +967,7 @@ gtk_source_view_populate_popup (GtkTextView *text_view,
 	GtkWidget *menu_item;
 
 	buffer = gtk_text_view_get_buffer (text_view);
-	if (!buffer && !GTK_IS_SOURCE_BUFFER (buffer))
+	if (!GTK_IS_SOURCE_BUFFER (buffer))
 		return;
 
 	/* separator */
@@ -1132,6 +1169,7 @@ gtk_source_view_get_lines (GtkTextView  *text_view,
 			   gint          first_y,
 			   gint          last_y,
 			   GArray       *buffer_coords,
+			   GArray       *line_heights,
 			   GArray       *numbers,
 			   gint         *countp)
 {
@@ -1142,6 +1180,8 @@ gtk_source_view_get_lines (GtkTextView  *text_view,
 
 	g_array_set_size (buffer_coords, 0);
 	g_array_set_size (numbers, 0);
+	if (line_heights != NULL)
+		g_array_set_size (line_heights, 0);
 
 	/* Get iter at first y */
 	gtk_text_view_get_line_at_y (text_view, &iter, first_y, NULL);
@@ -1158,6 +1198,8 @@ gtk_source_view_get_lines (GtkTextView  *text_view,
 		gtk_text_view_get_line_yrange (text_view, &iter, &y, &height);
 
 		g_array_append_val (buffer_coords, y);
+		if (line_heights)
+			g_array_append_val (line_heights, height);
 		last_line_num = gtk_text_iter_get_line (&iter);
 		g_array_append_val (numbers, last_line_num);
 
@@ -1181,6 +1223,8 @@ gtk_source_view_get_lines (GtkTextView  *text_view,
 		if (line_num != last_line_num)
 		{
 			g_array_append_val (buffer_coords, y);
+			if (line_heights)
+				g_array_append_val (line_heights, height);
 			g_array_append_val (numbers, line_num);
 			++count;
 		}
@@ -1190,7 +1234,7 @@ gtk_source_view_get_lines (GtkTextView  *text_view,
 }
 
 static gint
-sort_marks_by_priority (gconstpointer m1, 
+sort_marks_by_priority (gconstpointer m1,
 			gconstpointer m2,
 			gpointer data)
 {
@@ -1202,10 +1246,10 @@ sort_marks_by_priority (gconstpointer m1,
 	gint line2;
 
 	gtk_text_buffer_get_iter_at_mark (gtk_text_mark_get_buffer (GTK_TEXT_MARK (mark1)),
-					  &iter1, 
+					  &iter1,
 					  GTK_TEXT_MARK (mark1));
 	gtk_text_buffer_get_iter_at_mark (gtk_text_mark_get_buffer (GTK_TEXT_MARK (mark2)),
-					  &iter2, 
+					  &iter2,
 					  GTK_TEXT_MARK (mark2));
 
 	line1 = gtk_text_iter_get_line (&iter1);
@@ -1357,6 +1401,7 @@ gtk_source_view_paint_margin (GtkSourceView *view,
 				   y1,
 				   y2,
 				   pixels,
+				   NULL,
 				   numbers,
 				   &count);
 
@@ -1458,7 +1503,7 @@ gtk_source_view_paint_margin (GtkSourceView *view,
 					  layout);
 		}
 
-		if (view->priv->show_line_marks)
+		if (view->priv->show_line_marks && view->priv->source_buffer != NULL)
 		{
 			GSList *marks;
 
@@ -1481,6 +1526,368 @@ gtk_source_view_paint_margin (GtkSourceView *view,
 	g_object_unref (G_OBJECT (layout));
 }
 
+static void
+gtk_source_view_paint_line_background (GtkTextView    *text_view,
+				       GdkEventExpose *event,
+				       int             y, /* in buffer coordinates */
+				       int             height,
+				       const GdkColor *color)
+{
+	GdkRectangle visible_rect;
+	GdkRectangle line_rect;
+	gint win_y;
+	gint margin;
+	cairo_t *cr;
+
+	gtk_text_view_get_visible_rect (text_view, &visible_rect);
+
+	gtk_text_view_buffer_to_window_coords (text_view,
+					       GTK_TEXT_WINDOW_TEXT,
+					       visible_rect.x,
+					       y,
+					       &line_rect.x,
+					       &win_y);
+
+	line_rect.x = 0;
+	line_rect.width = visible_rect.width;
+	line_rect.y = win_y;
+	line_rect.height = height;
+
+	if (text_view->hadjustment)
+		margin = gtk_text_view_get_left_margin (text_view) -
+			 (int) text_view->hadjustment->value;
+	else
+		margin = gtk_text_view_get_left_margin (text_view);
+
+	line_rect.x += MAX (0, margin - 1);
+
+	cr = gdk_cairo_create (event->window);
+	gdk_cairo_set_source_color (cr, (GdkColor *)color);
+	cairo_set_line_width (cr, 1);
+	cairo_rectangle (cr, line_rect.x + .5, line_rect.y + .5,
+			 line_rect.width - 1, line_rect.height - 1);
+	cairo_stroke_preserve (cr);
+	cairo_fill (cr);
+}
+
+static void
+gtk_source_view_paint_marks_background (GtkSourceView  *view,
+					GdkEventExpose *event)
+{
+	GtkTextView *text_view;
+	GArray *numbers;
+	GArray *pixels;
+	GArray *heights;
+	gint y1, y2;
+	gint count;
+	gint i;
+
+	if (view->priv->source_buffer == NULL)
+		return;
+
+	text_view = GTK_TEXT_VIEW (view);
+
+	y1 = event->area.y;
+	y2 = y1 + event->area.height;
+
+	/* get the extents of the line printing */
+	gtk_text_view_window_to_buffer_coords (text_view,
+					       GTK_TEXT_WINDOW_TEXT,
+					       0,
+					       y1,
+					       NULL,
+					       &y1);
+
+	gtk_text_view_window_to_buffer_coords (text_view,
+					       GTK_TEXT_WINDOW_TEXT,
+					       0,
+					       y2,
+					       NULL,
+					       &y2);
+
+	numbers = g_array_new (FALSE, FALSE, sizeof (gint));
+	pixels = g_array_new (FALSE, FALSE, sizeof (gint));
+	heights = g_array_new (FALSE, FALSE, sizeof (gint));
+
+	/* get the line numbers and y coordinates. */
+	gtk_source_view_get_lines (text_view,
+				   y1,
+				   y2,
+				   pixels,
+				   heights,
+				   numbers,
+				   &count);
+
+	if (count == 0)
+	{
+		gint n = 0;
+		gint y;
+		gint height;
+		GtkTextIter iter;
+
+		gtk_text_buffer_get_start_iter (gtk_text_view_get_buffer (text_view), &iter);
+		gtk_text_view_get_line_yrange (text_view, &iter, &y, &height);
+
+		g_array_append_val (pixels, y);
+		g_array_append_val (pixels, height);
+		g_array_append_val (numbers, n);
+		count = 1;
+	}
+
+	DEBUG ({
+		g_message ("Painting marks background for line numbers %d - %d",
+			   g_array_index (numbers, gint, 0),
+			   g_array_index (numbers, gint, count - 1));
+	});
+
+	for (i = 0; i < count; ++i)
+	{
+		gint line_to_paint;
+		GSList *marks;
+		GdkColor *background;
+		int priority;
+
+		line_to_paint = g_array_index (numbers, gint, i);
+
+		marks = gtk_source_buffer_get_source_marks_at_line (view->priv->source_buffer,
+								    line_to_paint,
+								    NULL);
+
+		background = NULL;
+		priority = -1;
+
+		while (marks != NULL)
+		{
+			const gchar *category;
+			MarkCategory *cat = NULL;
+			category = gtk_source_mark_get_category (marks->data);
+			if (category != NULL)
+				cat = g_hash_table_lookup (view->priv->mark_categories, category);
+			if (cat != NULL && cat->background_set && cat->priority > priority)
+			{
+				background = &cat->background;
+				priority = cat->priority;
+			}
+			marks = g_slist_delete_link (marks, marks);
+		}
+
+		if (background != NULL)
+			gtk_source_view_paint_line_background (text_view, event,
+							       g_array_index (pixels, gint, i),
+							       g_array_index (heights, gint, i),
+							       background);
+	}
+
+	g_array_free (heights, TRUE);
+	g_array_free (pixels, TRUE);
+	g_array_free (numbers, TRUE);
+}
+
+static void
+draw_space_at_iter (cairo_t      *cr,
+		    GtkTextView  *view,
+		    GtkTextIter  *iter,
+		    GdkRectangle  rect)
+{
+	gint x, y;
+
+	gtk_text_view_buffer_to_window_coords (view,
+					       GTK_TEXT_WINDOW_TEXT,
+					       rect.x + rect.width / 2,
+					       rect.y + rect.height * 2 / 3,
+					       &x,
+					       &y);
+
+	cairo_save (cr);
+	cairo_move_to (cr, x, y);
+	cairo_arc (cr, x, y, 0.8, 0, 2 * G_PI);
+	cairo_restore (cr);
+}
+
+static void
+draw_tab_at_iter (cairo_t      *cr,
+		  GtkTextView  *view,
+		  GtkTextIter  *iter,
+		  GdkRectangle  rect)
+{
+	gint x, y;
+	double w, h;
+
+	gtk_text_view_buffer_to_window_coords (view,
+					       GTK_TEXT_WINDOW_TEXT,
+					       rect.x,
+					       rect.y + rect.height * 2 / 3,
+					       &x,
+					       &y);
+
+	w = rect.width;
+	h = rect.height;
+
+	cairo_save (cr);
+	cairo_move_to (cr, x + w * 1 / 8, y);
+	cairo_rel_line_to (cr, w * 6 / 8, 0);
+	cairo_rel_line_to (cr, -h * 1 / 4, -h * 1 / 4);
+	cairo_rel_move_to (cr, +h * 1 / 4, +h * 1 / 4);
+	cairo_rel_line_to (cr, -h * 1 / 4, +h * 1 / 4);
+	cairo_restore (cr);
+}
+
+static void
+draw_newline_at_iter (cairo_t      *cr,
+		      GtkTextView  *view,
+		      GtkTextIter  *iter,
+		      GdkRectangle  rect)
+{
+	gint x, y;
+	double w, h;
+
+	gtk_text_view_buffer_to_window_coords (view,
+					       GTK_TEXT_WINDOW_TEXT,
+					       rect.x,
+					       rect.y + rect.height * 1 / 3,
+					       &x,
+					       &y);
+
+	/* width for new line is 0, we use 2 * h */
+	w = 2 * rect.height;
+	h = rect.height;
+
+	cairo_save (cr);
+	if (gtk_widget_get_default_direction () == GTK_TEXT_DIR_LTR)
+	{
+		cairo_move_to (cr, x + w * 7 / 8, y);
+		cairo_rel_line_to (cr, 0, h * 1 / 3);
+		cairo_rel_line_to (cr, -w * 6 / 8, 0);
+		cairo_rel_line_to (cr, +h * 1 / 4, -h * 1 / 4);
+		cairo_rel_move_to (cr, -h * 1 / 4, +h * 1 / 4);
+		cairo_rel_line_to (cr, +h * 1 / 4, +h * 1 / 4);
+	}
+	else
+	{
+		cairo_move_to (cr, x + w * 1 / 8, y);
+		cairo_rel_line_to (cr, 0, h * 1 / 3);
+		cairo_rel_line_to (cr, w * 6 / 8, 0);
+		cairo_rel_line_to (cr, -h * 1 / 4, -h * 1 / 4);
+		cairo_rel_move_to (cr, +h * 1 / 4, +h * 1 / 4);
+		cairo_rel_line_to (cr, -h * 1 / 4, -h * 1 / 4);
+	}
+
+	cairo_restore (cr);
+}
+
+static void
+draw_spaces_at_iter (cairo_t       *cr,
+		     GtkSourceView *view,
+		     GtkTextIter   *iter,
+		     GdkRectangle   rect)
+{
+	gunichar c;
+
+	c = gtk_text_iter_get_char (iter);
+
+	if (view->priv->draw_spaces & GTK_SOURCE_DRAW_SPACES_TAB &&
+	    c == '\t')
+	{
+		draw_tab_at_iter (cr, GTK_TEXT_VIEW (view), iter, rect);
+	}
+	else if (view->priv->draw_spaces & GTK_SOURCE_DRAW_SPACES_SPACE &&
+	         g_unichar_type (c) == G_UNICODE_SPACE_SEPARATOR)
+	{
+		draw_space_at_iter (cr, GTK_TEXT_VIEW (view), iter, rect);
+	}
+	else if (view->priv->draw_spaces & GTK_SOURCE_DRAW_SPACES_NEWLINE &&
+		 gtk_text_iter_ends_line (iter))
+	{
+		draw_newline_at_iter (cr, GTK_TEXT_VIEW (view), iter, rect);
+	}
+}
+
+static void
+draw_tabs_and_spaces (GtkSourceView  *view,
+		      GdkEventExpose *event)
+{
+	GtkTextView *text_view;
+	gint x1, y1, x2, y2;
+	GtkTextIter s, e;
+	cairo_t *cr;
+
+	text_view = GTK_TEXT_VIEW (view);
+
+	x1 = event->area.x;
+	y1 = event->area.y;
+	x2 = x1 + event->area.width;
+	y2 = y1 + event->area.height;
+
+	gtk_text_view_window_to_buffer_coords (text_view,
+					       GTK_TEXT_WINDOW_TEXT,
+					       x1,
+					       y1,
+					       &x1,
+					       &y1);
+
+	gtk_text_view_window_to_buffer_coords (text_view,
+					       GTK_TEXT_WINDOW_TEXT,
+					       x2,
+					       y2,
+					       &x2,
+					       &y2);
+
+	gtk_text_view_get_iter_at_location  (text_view,
+                                             &s,
+                                             x1, y1);
+	gtk_text_view_get_iter_at_location  (text_view,
+                                             &e,
+                                             x2, y2);
+
+	cr = gdk_cairo_create (event->window);
+
+	cairo_set_source_rgba (cr,
+			       view->priv->spaces_color->red / 65535.,
+			       view->priv->spaces_color->green / 65535.,
+			       view->priv->spaces_color->blue / 65535.,
+			       1);
+	cairo_set_line_width (cr, 0.8);
+
+	do {
+		GdkRectangle rect;
+		gint ly;
+
+		gtk_text_view_get_iter_location (text_view, &s, &rect);
+
+		/* just iterate on the text that is in the exposed area */
+		if (rect.x > x2)
+		{
+			if (!gtk_text_iter_forward_line	(&s))
+				break;
+
+			/* move to the first iter in the exposed area of
+			 * the next line */
+			gtk_text_view_get_line_yrange (text_view, &s, &ly, NULL);
+
+			gtk_text_view_get_iter_at_location  (text_view,
+							     &s,
+							     x1, ly);
+
+			/* move back one char otherwise tabs may not
+			 * be redrawn */
+			if (!gtk_text_iter_starts_line (&s))
+				gtk_text_iter_backward_char (&s);
+
+			continue;
+		}
+
+		draw_spaces_at_iter (cr, view, &s, rect);
+
+		if (!gtk_text_iter_forward_char (&s))
+			break;
+
+	} while (gtk_text_iter_compare (&s, &e) <= 0);
+
+	cairo_stroke (cr);
+
+	cairo_destroy (cr);
+}
+
 static gint
 gtk_source_view_expose (GtkWidget      *widget,
 			GdkEventExpose *event)
@@ -1552,67 +1959,29 @@ gtk_source_view_expose (GtkWidget      *widget,
 				gdk_window_invalidate_rect (w, NULL, FALSE);
 		}
 
-		if (view->priv->highlight_current_line &&
+		if (GTK_WIDGET_IS_SENSITIVE(view) && view->priv->highlight_current_line &&
 		    (event->window == gtk_text_view_get_window (text_view, GTK_TEXT_WINDOW_TEXT)))
 		{
-			GdkRectangle visible_rect;
-			GdkRectangle redraw_rect;
 			GtkTextIter cur;
-			gint y;
-			gint height;
-			gint win_y;
-			GdkGC *gc;
-			gint margin;
+			gint y, height;
+			GdkColor *color;
 
 			gtk_text_buffer_get_iter_at_mark (text_view->buffer,
 							  &cur,
 							  gtk_text_buffer_get_insert (text_view->buffer));
-
 			gtk_text_view_get_line_yrange (text_view, &cur, &y, &height);
 
-			gtk_text_view_get_visible_rect (text_view, &visible_rect);
-
-			gtk_text_view_buffer_to_window_coords (text_view,
-						       GTK_TEXT_WINDOW_TEXT,
-						       visible_rect.x,
-						       visible_rect.y,
-						       &redraw_rect.x,
-						       &redraw_rect.y);
-
-			gtk_text_view_buffer_to_window_coords (text_view,
-						       GTK_TEXT_WINDOW_TEXT,
-						       0,
-						       y,
-						       NULL,
-						       &win_y);
-
-			redraw_rect.width = visible_rect.width;
-			redraw_rect.height = visible_rect.height;
-
-			if (view->priv->current_line_gc)
-				gc = view->priv->current_line_gc;
+			if (view->priv->current_line_color_set)
+				color = &view->priv->current_line_color;
 			else
-				gc = widget->style->bg_gc[GTK_WIDGET_STATE (widget)];
+				color = &widget->style->bg[GTK_WIDGET_STATE (widget)];
 
-			if (text_view->hadjustment)
-			{
-				margin = gtk_text_view_get_left_margin (text_view) -
-					 (int) text_view->hadjustment->value;
-			}
-			else
-			{
-				margin = gtk_text_view_get_left_margin (text_view);
-			}
-
-			gdk_draw_rectangle (event->window,
-					    gc,
-					    TRUE,
-					    redraw_rect.x + MAX (0, margin - 1),
-					    win_y,
-					    redraw_rect.width,
-					    height);
+			gtk_source_view_paint_line_background (text_view, event, y, height, color);
 		}
 
+		if (event->window == gtk_text_view_get_window (text_view, GTK_TEXT_WINDOW_TEXT))
+			gtk_source_view_paint_marks_background (view, event);
+
 		/* Have GtkTextView draw the text first. */
 		if (GTK_WIDGET_CLASS (gtk_source_view_parent_class)->expose_event)
 			event_handled =
@@ -1631,7 +2000,7 @@ gtk_source_view_expose (GtkWidget      *widget,
 			static GTimer *timer = NULL;
 #endif
 
-			g_return_val_if_fail (view->priv->right_margin_line_color != NULL, 
+			g_return_val_if_fail (view->priv->right_margin_line_color != NULL,
 					      event_handled);
 
 			if (view->priv->cached_right_margin_pos < 0)
@@ -1689,7 +2058,7 @@ gtk_source_view_expose (GtkWidget      *widget,
 
 			/* Only draw the overlay when the style scheme explicitly sets it. */
 			if (view->priv->right_margin_overlay_color != NULL)
-			{	
+			{
 				/* Draw the rectangle next to the line (x+.5). */
 				cairo_rectangle (cr,
 						 x + .5,
@@ -1714,6 +2083,12 @@ gtk_source_view_expose (GtkWidget      *widget,
 				           g_timer_elapsed (timer, NULL) * 1000);
 			});
 		}
+
+		if (view->priv->draw_spaces != 0 &&
+		    (event->window == gtk_text_view_get_window (text_view, GTK_TEXT_WINDOW_TEXT)))
+		{
+			draw_tabs_and_spaces (view, event);
+		}
 	}
 
 	DEBUG ({
@@ -1958,6 +2333,7 @@ set_tab_stops_internal (GtkSourceView *view)
 
 	gtk_text_view_set_tabs (GTK_TEXT_VIEW (view),
 				tab_array);
+	view->priv->tabs_set = TRUE;
 
 	pango_tab_array_free (tab_array);
 
@@ -2054,6 +2430,20 @@ gtk_source_view_get_indent_width (GtkSourceView *view)
 	return view->priv->indent_width;
 }
 
+static void
+mark_category_set_background (MarkCategory *cat, const GdkColor *background)
+{
+	if (background != NULL)
+	{
+		cat->background_set = TRUE;
+		cat->background = *background;
+	}
+	else
+	{
+		cat->background_set = FALSE;
+	}
+}
+
 static MarkCategory *
 mark_category_new (gint priority, GdkPixbuf *pixbuf)
 {
@@ -2147,6 +2537,9 @@ gtk_source_view_set_mark_category_pixbuf (GtkSourceView *view,
 			cat->pixbuf = NULL;
 		}
 	}
+
+	/* We may need to redraw the margin now */
+	gtk_widget_queue_draw (GTK_WIDGET (view));
 }
 
 /**
@@ -2176,6 +2569,80 @@ gtk_source_view_get_mark_category_pixbuf (GtkSourceView *view,
 		return NULL;
 }
 
+/**
+ * gtk_source_view_set_mark_category_background:
+ * @view: a #GtkSourceView.
+ * @category: a mark category.
+ * @color: background color or %NULL to unset it.
+ *
+ * Sets given background @color for mark @category.
+ * If @color is #NULL, the background color is unset.
+ *
+ * Since: 2.4
+ */
+void
+gtk_source_view_set_mark_category_background (GtkSourceView  *view,
+					      const gchar    *category,
+					      const GdkColor *color)
+{
+	MarkCategory *cat;
+
+	g_return_if_fail (GTK_IS_SOURCE_VIEW (view));
+	g_return_if_fail (category != NULL);
+
+	cat = g_hash_table_lookup (view->priv->mark_categories, category);
+
+	if (color != NULL && cat == NULL)
+	{
+		cat = mark_category_new (0, NULL);
+		g_hash_table_insert (view->priv->mark_categories,
+				     g_strdup (category),
+				     cat);
+	}
+
+	if (cat != NULL)
+		mark_category_set_background (cat, color);
+
+	/* We may need to redraw the text background now */
+	gtk_widget_queue_draw (GTK_WIDGET (view));
+}
+
+/**
+ * gtk_source_view_get_mark_category_background:
+ * @view: a #GtkSourceView.
+ * @category: a mark category.
+ * @dest: destination #GdkColor structure to fill in.
+ *
+ * Gets the background color associated with given @category.
+ *
+ * Return value: %TRUE if background color for @category was set
+ * and @dest is set to a valid color, or %FALSE otherwise.
+ *
+ * Since: 2.4
+ */
+gboolean
+gtk_source_view_get_mark_category_background (GtkSourceView *view,
+					      const gchar   *category,
+					      GdkColor      *dest)
+{
+	MarkCategory *cat;
+
+	g_return_val_if_fail (GTK_IS_SOURCE_VIEW (view), FALSE);
+	g_return_val_if_fail (category != NULL, FALSE);
+	g_return_val_if_fail (dest != NULL, FALSE);
+
+	cat = g_hash_table_lookup (view->priv->mark_categories, category);
+	if (cat != NULL && cat->background_set)
+	{
+		*dest = cat->background;
+		return TRUE;
+	}
+	else
+	{
+		return FALSE;
+	}
+}
+
 /**
  * gtk_source_view_set_mark_category_priority:
  * @view: a #GtkSourceView.
@@ -2185,7 +2652,7 @@ gtk_source_view_get_mark_category_pixbuf (GtkSourceView *view,
  * Set the @priority for the given mark @category. When there are
  * multiple marks on the same line, marks of categories with
  * higher priorities will be drawn on top.
- * 
+ *
  * Since: 2.2
  */
 void
@@ -2208,6 +2675,9 @@ gtk_source_view_set_mark_category_priority (GtkSourceView *view,
 	}
 	else
 		cat->priority = priority;
+
+	/* We may need to redraw now, if the priorities affect appearance */
+	gtk_widget_queue_draw (GTK_WIDGET (view));
 }
 
 /**
@@ -2334,7 +2804,7 @@ indent_lines (GtkSourceView *view, GtkTextIter *start, GtkTextIter *end)
 
 		indent_width = get_real_indent_width (view);
 		spaces = indent_width % view->priv->tab_width;
-		tabs = indent_width / view->priv->tab_width;		
+		tabs = indent_width / view->priv->tab_width;
 
 		tab_buffer = get_indent_string (tabs, spaces);
 	}
@@ -2677,10 +3147,13 @@ gtk_source_view_key_press_event (GtkWidget   *widget,
 	GtkTextMark *mark;
 	guint modifiers;
 	gint key;
+	gboolean editable;
 
 	view = GTK_SOURCE_VIEW (widget);
 	buf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (widget));
 
+	editable = gtk_text_view_get_editable (GTK_TEXT_VIEW (widget));
+
 	/* Be careful when testing for modifier state equality:
 	 * caps lock, num lock,etc need to be taken into account */
 	modifiers = gtk_accelerator_get_default_mod_mask ();
@@ -2729,11 +3202,12 @@ gtk_source_view_key_press_event (GtkWidget   *widget,
 	}
 
 	/* if tab or shift+tab:
-	 * with shift+tab key is GDK_ISO_Left_Tab (depends on X?)
+	 * with shift+tab key is GDK_ISO_Left_Tab (yay! on win32 and mac too!)
 	 */
 	if ((key == GDK_Tab || key == GDK_KP_Tab || key == GDK_ISO_Left_Tab) &&
 	    ((event->state & modifiers) == 0 ||
-	     (event->state & modifiers) == GDK_SHIFT_MASK))
+	     (event->state & modifiers) == GDK_SHIFT_MASK) &&
+	    editable)
 	{
 		GtkTextIter s, e;
 		gboolean has_selection;
@@ -2768,7 +3242,8 @@ gtk_source_view_key_press_event (GtkWidget   *widget,
 
 	/* Alt+up/down moves the lines */
 	if ((key == GDK_Up || key == GDK_Down) &&
-	    ((event->state & modifiers) == GDK_MOD1_MASK))
+	    ((event->state & modifiers) == GDK_MOD1_MASK) &&
+	    editable)
 	{
 		move_lines (GTK_TEXT_VIEW (widget), key == GDK_Down);
 		return TRUE;
@@ -3225,6 +3700,48 @@ gtk_source_view_get_smart_home_end (GtkSourceView *view)
 	return view->priv->smart_home_end;
 }
 
+/**
+ * gtk_source_view_set_draw_spaces:
+ * @view: a #GtkSourceView.
+ * @flags: #GtkSourceDrawSpacesFlags specifing how white spaces should
+ * be displayed
+ *
+ * Set if and how the spaces should be visualized. Specifying @flags as 0 will
+ * disable display of spaces.
+ **/
+void
+gtk_source_view_set_draw_spaces (GtkSourceView            *view,
+				 GtkSourceDrawSpacesFlags  flags)
+{
+	g_return_if_fail (GTK_IS_SOURCE_VIEW (view));
+
+	if (view->priv->draw_spaces == flags)
+		return;
+
+	view->priv->draw_spaces = flags;
+
+	gtk_widget_queue_draw (GTK_WIDGET (view));
+
+	g_object_notify (G_OBJECT (view), "draw-spaces");
+}
+
+/**
+ * gtk_source_view_get_draw_spaces:
+ * @view: a #GtkSourceView
+ *
+ * Returns the #GtkSourceDrawSpacesFlags specifying if and how spaces
+ * should be displayed for this @view.
+ * 
+ * Returns: the #GtkSourceDrawSpacesFlags, 0 if no spaces should be drawn.
+ **/
+GtkSourceDrawSpacesFlags
+gtk_source_view_get_draw_spaces (GtkSourceView *view)
+{
+	g_return_val_if_fail (GTK_IS_SOURCE_VIEW (view), 0);
+
+	return view->priv->draw_spaces;
+}
+
 /**
  * gtk_source_view_style_set:
  * @widget: a #GtkSourceView.
@@ -3250,8 +3767,10 @@ gtk_source_view_style_set (GtkWidget *widget, GtkStyle *previous_style)
 		 * emission and we can't set the tab array since the
 		 * text view doesn't have a default style yet */
 
-		/* re-set tab stops */
-		set_tab_stops_internal (view);
+		/* re-set tab stops, but only if we already modified them, i.e.
+		 * do nothing with good old 8-space tabs */
+		if (view->priv->tabs_set)
+			set_tab_stops_internal (view);
 
 		/* make sure the margin position is recalculated on next expose */
 		view->priv->cached_right_margin_pos = -1;
@@ -3259,31 +3778,11 @@ gtk_source_view_style_set (GtkWidget *widget, GtkStyle *previous_style)
 }
 
 static void
-update_current_line_gc (GtkSourceView *view)
+update_current_line_color (GtkSourceView *view)
 {
-	GdkColor color;
-	GtkWidget *widget = GTK_WIDGET (view);
-
-	if (!GTK_WIDGET_REALIZED (view))
-		return;
-
-	if (view->priv->current_line_gc)
-	{
-		g_object_unref (view->priv->current_line_gc);
-		view->priv->current_line_gc = NULL;
-	}
-
-	if (view->priv->style_scheme &&
-	    _gtk_source_style_scheme_get_current_line_color (view->priv->style_scheme, &color))
-	{
-		GdkGCValues values;
-		gdk_colormap_alloc_color (gtk_widget_get_colormap (widget),
-					  &color, TRUE, TRUE);
-		values.foreground = color;
-		view->priv->current_line_gc = gdk_gc_new_with_values (widget->window,
-								      &values,
-								      GDK_GC_FOREGROUND);
-	}
+	view->priv->current_line_color_set =
+		_gtk_source_style_scheme_get_current_line_color (view->priv->style_scheme,
+								 &view->priv->current_line_color);
 }
 
 static void
@@ -3306,7 +3805,7 @@ update_right_margin_colors (GtkSourceView *view)
 		view->priv->right_margin_overlay_color = NULL;
 	}
 
-	if (view->priv->style_scheme) 
+	if (view->priv->style_scheme)
 	{
 		GtkSourceStyle	*style;
 
@@ -3317,8 +3816,8 @@ update_right_margin_colors (GtkSourceView *view)
 			gchar *color_str = NULL;
 			gboolean color_set;
 			GdkColor color;
-			
-			g_object_get (style, 
+
+			g_object_get (style,
 				      "foreground-set", &color_set,
 				      "foreground", &color_str,
 				      NULL);
@@ -3327,11 +3826,11 @@ update_right_margin_colors (GtkSourceView *view)
 			{
 				view->priv->right_margin_line_color = gdk_color_copy (&color);
 			}
-			
+
 			g_free (color_str);
 			color_str = NULL;
-			
-			g_object_get (style, 
+
+			g_object_get (style,
 				      "background-set", &color_set,
 				      "background", &color_str,
 				      NULL);
@@ -3344,11 +3843,53 @@ update_right_margin_colors (GtkSourceView *view)
 			g_free (color_str);
 		}
 	}
-	
+
 	if (view->priv->right_margin_line_color == NULL)
 		view->priv->right_margin_line_color = gdk_color_copy (&widget->style->text[GTK_STATE_NORMAL]);
 }
 
+static void
+update_spaces_color (GtkSourceView *view)
+{
+	GtkWidget *widget = GTK_WIDGET (view);
+
+	if (!GTK_WIDGET_REALIZED (view))
+		return;
+
+	if (view->priv->spaces_color != NULL)
+	{
+		gdk_color_free (view->priv->spaces_color);
+		view->priv->spaces_color = NULL;
+	}
+
+	if (view->priv->style_scheme)
+	{
+		GtkSourceStyle	*style;
+
+		style = _gtk_source_style_scheme_get_draw_spaces_style (view->priv->style_scheme);
+
+		if (style != NULL)
+		{
+			gchar *color_str = NULL;
+			GdkColor color;
+
+			g_object_get (style,
+				      "foreground", &color_str,
+				      NULL);
+
+			if (color_str != NULL && gdk_color_parse (color_str, &color))
+			{
+				view->priv->spaces_color = gdk_color_copy (&color);
+			}
+
+			g_free (color_str);
+		}
+	}
+
+	if (view->priv->spaces_color == NULL)
+		view->priv->spaces_color = gdk_color_copy (&widget->style->text[GTK_STATE_INSENSITIVE]);
+}
+
 static void
 gtk_source_view_realize (GtkWidget *widget)
 {
@@ -3362,20 +3903,9 @@ gtk_source_view_realize (GtkWidget *widget)
 		view->priv->style_scheme_applied = TRUE;
 	}
 
-	update_current_line_gc (view);
+	update_current_line_color (view);
 	update_right_margin_colors (view);
-}
-
-static void
-gtk_source_view_unrealize (GtkWidget *widget)
-{
-	GtkSourceView *view = GTK_SOURCE_VIEW (widget);
-
-	if (view->priv->current_line_gc)
-		g_object_unref (view->priv->current_line_gc);
-	view->priv->current_line_gc = NULL;
-
-	GTK_WIDGET_CLASS (gtk_source_view_parent_class)->unrealize (widget);
+	update_spaces_color (view);
 }
 
 static void
@@ -3402,8 +3932,9 @@ gtk_source_view_update_style_scheme (GtkSourceView *view)
 		if (GTK_WIDGET_REALIZED (view))
 		{
 			_gtk_source_style_scheme_apply (new_scheme, GTK_WIDGET (view));
-			update_current_line_gc (view);
+			update_current_line_color (view);
 			update_right_margin_colors (view);
+			update_spaces_color (view);
 			view->priv->style_scheme_applied = TRUE;
 		}
 		else
